{#
    warning.html
    ~~~~~~~~~~~

    A simple page to emit a warning when no docstring text was found for the
    one a user was looking for

    :copyright: Copyright (c) Spyder Project Contributors
    :license: MIT license
#}

<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
  "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">

<html>

<head>
    <meta http-equiv="Content-Type" content="text/html; charset=utf-8" />
    <link rel="stylesheet" href="file:///{{css_path}}/default.css" type="text/css" />
</head>

<body>
  <div id="doc-warning">{{text}}</div>
</body>

</html>
