<!--
    kernel_error.html
    ~~~~~~~~~~~~~~~~~

    Kernel error template for the IPython Console

    :copyright: Copyright by the Spyder Project Contributors
    :license: MIT license
-->

<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
  "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">

<html>

<head>
    <meta http-equiv="Content-Type" content="text/html; charset=utf-8"/>
    <link rel="stylesheet" href="file:///${css_path}/default.css" type="text/css"/>
</head>

<body>
  <div class="panel panel-danger">
    <div class="panel-heading">
      <div class="panel-title">${message}</div>
    </div>
    <div class="panel-body">
      ${error}
    </div>
  </div>
</body>

</html>
